use core::fmt::Display;

use arrayvec::ArrayString;
use crc::{Crc, CRC_32_ISCSI};
use embedded_storage::nor_flash::{NorFlash, ReadNorFlash};
use stm32f4xx_hal::flash::{FlashExt, LockedFlash};

use crate::geo::distance_km;

const CASTAGNOLI: Crc<u32> = Crc::<u32>::new(&CRC_32_ISCSI);

#[derive(Clone)]
pub struct Config {
    pub callsign: ArrayString<252>,
    pub icon: u16,
    pub ssid: u8,
    pub max_hops: u8,
    pub comment: ArrayString<512>,
    pub transmit_period_seconds: u64,
    pub gps: GpsSetting,
    pub enable_transmit: bool,
    pub enable_digipeating: bool,
    pub black_hole: MaybeBlackHole,
    pub antenna_height: Maybe<u8>,
    pub antenna_gain: Maybe<f32>,
    pub tx_power: Maybe<f32>,
}

impl Config {
    pub fn save_to_flash(&self, flash: &mut LockedFlash) {
        let mut unlocked = flash.unlocked();

        // Erase last flash sector (128kB)
        NorFlash::erase(&mut unlocked, 384 * 1024, 512 * 1024).unwrap();

        // Write our config
        let mut buf = [0; 839];
        self.serialize(&mut buf);
        NorFlash::write(&mut unlocked, 384 * 1024, &buf).unwrap();
    }

    pub fn load_from_flash(flash: &mut LockedFlash) -> Option<Self> {
        let mut buf = [0; 839];
        ReadNorFlash::read(flash, 384 * 1024, &mut buf).unwrap();

        Self::deserialize(&buf)
    }

    fn serialize(&self, buf: &mut [u8; 839]) {
        // version
        // allows us to update the config structure in the future
        // while maintaining backwards compatibility
        buf[0] = 0;

        buf[1..3].copy_from_slice(&self.icon.to_le_bytes());
        buf[3] = self.ssid;
        buf[4] = self.max_hops;
        buf[5..13].copy_from_slice(&self.transmit_period_seconds.to_le_bytes());
        match self.gps {
            GpsSetting::Disabled => {
                buf[13] = 0;
            }

            GpsSetting::Fixed(lat, lon) => {
                buf[13] = 1;
                buf[14..22].copy_from_slice(&lat.to_le_bytes());
                buf[22..30].copy_from_slice(&lon.to_le_bytes());
            }

            GpsSetting::Receiver => {
                buf[13] = 2;
            }
        }

        let enable_tx: u8 = self.enable_transmit.into();
        let enable_digi: u8 = self.enable_digipeating.into();
        buf[30] = enable_tx | (enable_digi << 1);

        buf[31] = self.callsign.len().try_into().unwrap();
        buf[32..(32 + self.callsign.len())].copy_from_slice(self.callsign.as_bytes());
        buf[284..286].copy_from_slice(&u16::try_from(self.comment.len()).unwrap().to_le_bytes());
        buf[286..(286 + self.comment.len())].copy_from_slice(self.comment.as_bytes());

        self.black_hole
            .encode((&mut buf[798..823]).try_into().unwrap());

        match self.antenna_height {
            Maybe::Some(height) => {
                buf[823] = 1;
                buf[824] = height;
            }
            Maybe::None => buf[823] = 0,
        }

        match self.antenna_gain {
            Maybe::Some(gain) => {
                buf[825] = 1;
                buf[826..830].copy_from_slice(&gain.to_le_bytes());
            }
            Maybe::None => buf[825] = 0,
        }

        match self.tx_power {
            Maybe::Some(power) => {
                buf[830] = 1;
                buf[831..835].copy_from_slice(&power.to_le_bytes());
            }
            Maybe::None => {
                buf[830] = 0;
            }
        }

        // Checksum
        let checksum = CASTAGNOLI.checksum(&buf[0..835]);
        buf[835..839].copy_from_slice(&checksum.to_le_bytes());
    }

    fn deserialize(buf: &[u8; 839]) -> Option<Self> {
        match buf[0] {
            0 => Self::deserialize_v0(buf),
            _ => None,
        }
    }

    fn deserialize_v0(buf: &[u8; 839]) -> Option<Self> {
        let expected_checksum = CASTAGNOLI.checksum(&buf[0..835]);
        let actual_checksum = u32::from_le_bytes(buf[835..839].try_into().unwrap());
        if expected_checksum != actual_checksum {
            return None;
        }

        let icon = u16::from_le_bytes(buf[1..3].try_into().unwrap());
        let ssid = buf[3];
        let max_hops = buf[4];
        let transmit_period_seconds = u64::from_le_bytes(buf[5..13].try_into().unwrap());
        let gps = match buf[13] {
            0 => GpsSetting::Disabled,
            1 => {
                let lat = f64::from_le_bytes(buf[14..22].try_into().unwrap());
                let lon = f64::from_le_bytes(buf[22..30].try_into().unwrap());

                GpsSetting::Fixed(lat, lon)
            }
            2 => GpsSetting::Receiver,
            _ => return None,
        };

        let enable_transmit = (buf[30] & 1) > 0;
        let enable_digipeating = (buf[30] & 2) > 0;

        let callsign_len: usize = buf[31].into();
        let callsign = core::str::from_utf8(&buf[32..(32 + callsign_len)]).ok()?;
        let callsign = ArrayString::from(callsign).ok()?;

        let comment_len: usize = u16::from_le_bytes(buf[284..286].try_into().unwrap()).into();
        if comment_len > 512 {
            return None;
        }
        let comment = core::str::from_utf8(&buf[286..(286 + comment_len)]).ok()?;
        let comment = ArrayString::from(comment).ok()?;

        let black_hole = MaybeBlackHole::decode(&buf[798..823].try_into().unwrap())?;

        let antenna_height = if buf[823] > 0 {
            Maybe::Some(buf[824])
        } else {
            Maybe::None
        };

        let antenna_gain = if buf[825] > 0 {
            Maybe::Some(f32::from_le_bytes(buf[826..830].try_into().unwrap()))
        } else {
            Maybe::None
        };

        let tx_power = if buf[830] > 0 {
            Maybe::Some(f32::from_le_bytes(buf[831..835].try_into().unwrap()))
        } else {
            Maybe::None
        };

        Some(Config {
            callsign,
            icon,
            ssid,
            max_hops,
            comment,
            transmit_period_seconds,
            gps,
            enable_transmit,
            enable_digipeating,
            black_hole,
            antenna_height,
            antenna_gain,
            tx_power,
        })
    }
}

impl Default for Config {
    fn default() -> Self {
        Config {
            callsign: "N0CALL".try_into().unwrap(),
            icon: 0,
            ssid: 0,
            max_hops: 2,
            comment: "".try_into().unwrap(),
            transmit_period_seconds: 60,
            gps: GpsSetting::Disabled,
            enable_transmit: false,
            enable_digipeating: true,
            black_hole: MaybeBlackHole::None,
            antenna_height: Maybe::None,
            antenna_gain: Maybe::None,
            tx_power: Maybe::Some(30.0),
        }
    }
}

#[derive(Copy, Clone)]
pub enum GpsSetting {
    Disabled,
    Fixed(f64, f64),
    Receiver,
}

impl Display for GpsSetting {
    fn fmt(&self, f: &mut core::fmt::Formatter<'_>) -> core::fmt::Result {
        match self {
            GpsSetting::Disabled => write!(f, "disabled"),
            GpsSetting::Fixed(lat, lon) => write!(f, "fixed {lat:.6} {lon:.6}"),
            GpsSetting::Receiver => write!(f, "receiver"),
        }
    }
}

#[derive(Copy, Clone)]
pub enum MaybeBlackHole {
    Some(BlackHoleSetting),
    None,
}

impl MaybeBlackHole {
    pub fn in_black_hole(&self, lat: f64, lon: f64) -> bool {
        match self {
            MaybeBlackHole::Some(bhs) => {
                let dist_km = distance_km(lat, lon, bhs.latitude, bhs.longitude);

                dist_km * 1000.0 < bhs.radius_meters
            }
            MaybeBlackHole::None => false,
        }
    }

    fn encode(&self, buf: &mut [u8; 25]) {
        match self {
            Self::None => {
                buf[0] = 0;
            }

            Self::Some(bhs) => {
                buf[0] = 1;
                bhs.encode((&mut buf[1..]).try_into().unwrap());
            }
        }
    }

    fn decode(buf: &[u8; 25]) -> Option<Self> {
        match buf[0] {
            0 => Some(Self::None),
            1 => Some(MaybeBlackHole::Some(BlackHoleSetting::decode(
                buf[1..].try_into().unwrap(),
            ))),
            _ => None,
        }
    }
}

impl Display for MaybeBlackHole {
    fn fmt(&self, f: &mut core::fmt::Formatter<'_>) -> core::fmt::Result {
        match self {
            MaybeBlackHole::Some(bhs) => write!(f, "{bhs}"),
            MaybeBlackHole::None => write!(f, "<None>"),
        }
    }
}

#[derive(Copy, Clone)]
pub struct BlackHoleSetting {
    pub latitude: f64,
    pub longitude: f64,
    pub radius_meters: f64,
}

impl BlackHoleSetting {
    fn encode(&self, buf: &mut [u8; 24]) {
        buf[0..8].copy_from_slice(&self.latitude.to_le_bytes());
        buf[8..16].copy_from_slice(&self.longitude.to_le_bytes());
        buf[16..24].copy_from_slice(&self.radius_meters.to_le_bytes());
    }

    fn decode(buf: &[u8; 24]) -> Self {
        let latitude = f64::from_le_bytes(buf[0..8].try_into().unwrap());
        let longitude = f64::from_le_bytes(buf[8..16].try_into().unwrap());
        let radius_meters = f64::from_le_bytes(buf[16..24].try_into().unwrap());

        Self {
            latitude,
            longitude,
            radius_meters,
        }
    }
}

impl Display for BlackHoleSetting {
    fn fmt(&self, f: &mut core::fmt::Formatter<'_>) -> core::fmt::Result {
        write!(
            f,
            "{:.6} {:.6} {:.2}",
            self.latitude, self.longitude, self.radius_meters
        )
    }
}

// Like Option, but local, so we can implement Display on it
#[derive(Copy, Clone)]
pub enum Maybe<T> {
    Some(T),
    None,
}

impl<T> Display for Maybe<T>
where
    T: Display,
{
    fn fmt(&self, f: &mut core::fmt::Formatter<'_>) -> core::fmt::Result {
        match self {
            Maybe::Some(x) => write!(f, "{x}"),
            Maybe::None => write!(f, "<None>"),
        }
    }
}
